require 'spec_helper'

module Contender

  describe CountdownLatch do
    it 'requires a positive initial count' do
      expect {
        CountdownLatch.new -1
      }.to raise_error ArgumentError
    end

    describe 'counting down from 1' do
      before do
        @latch = CountdownLatch.new 1
        @name = :foo
      end

      it 'blocks until counted down in another thread' do
        Thread.new do
          @name = :bar
          @latch.countdown
        end

        @latch.await.should == true
        @latch.count.should == 0

        @name.should == :bar
      end

      it 'blocks another thread until counted down' do
        Thread.new do
          @latch.await.should == true
          @latch.count.should == 0
          @name.should == :bar
        end

        @name = :bar
        @latch.countdown
      end

      it 'returns false if timed out' do
        @latch.await(0).should == false
      end
    end

    describe 'counting down from 0' do
      before do
        @latch = CountdownLatch.new 0
      end

      it 'does not block' do
        @latch.await.should == true
        @latch.count.should == 0
      end
    end

    describe 'counting down from 2' do
      before do
        @latch = CountdownLatch.new 2
        @name = :foo
      end

      it 'blocks until counted down in 1 parallel thread' do
        Thread.new do
          @latch.countdown
          @name = :bar
          @latch.countdown
        end

        @latch.await
        @latch.count.should == 0
        @name.should == :bar
      end

      it 'blocks until counted down in 2 parallel threads' do
        Thread.new do
          @latch.countdown
        end

        Thread.new do
          @name = :bar
          @latch.countdown
        end

        @latch.await
        @latch.count.should == 0
        @name.should == :bar
      end

      it 'blocks until counted down in 2 chained threads' do
        Thread.new do
          @latch.countdown

          Thread.new do
            @name = :bar
            @latch.countdown
          end
        end

        @latch.await
        @latch.count.should == 0
        @name.should == :bar
      end
    end

    describe 'with interleaved latches' do
      before do
        @check_latch = CountdownLatch.new 1
        @change_latch1 = CountdownLatch.new 1
        @change_latch2 = CountdownLatch.new 1
        @name = :foo
      end

      it 'blocks the correct thread' do
        Thread.new do
          @name = :bar
          @change_latch1.countdown

          @check_latch.await

          @name = :baz
          @change_latch2.countdown
        end

        @change_latch1.await
        @name.should == :bar

        @check_latch.countdown

        @change_latch2.await
        @name.should == :baz
      end
    end
  end

end
